from gevent import monkey

monkey.patch_all()

import pytest
import gevent
import marshmallow
from channelstream.server_state import get_state
from channelstream.channel import Channel


@pytest.mark.usefixtures("cleanup_globals", "pyramid_config")
class TestConnectViews(object):
    def test_bad_json(self, dummy_request, test_uuids):
        from channelstream.wsgi_views.server import connect

        dummy_request.json_body = {}
        try:
            connect(dummy_request)
        except marshmallow.exceptions.ValidationError as exc:
            assert exc.messages == {"username": ["Missing data for required field."]}

    def test_good_json(self, dummy_request, test_uuids):
        server_state = get_state()
        from channelstream.wsgi_views.server import connect

        dummy_request.json_body = {
            "username": "username",
            "conn_id": str(test_uuids[1]),
            "fresh_user_state": {"key": "foo"},
            "user_state": {"bar": "baz"},
            "state_public_keys": ["bar"],
            "channels": ["a", "aB"],
            "channel_configs": {"a": {"store_history": True, "history_size": 2}},
        }
        assert server_state.channels == {}
        result = connect(dummy_request)
        assert len(server_state.channels.keys()) == 2
        assert "username" in server_state.users
        assert test_uuids[1] in server_state.connections
        assert result["channels"] == ["a", "aB"]
        assert result["state"] == {"bar": "baz", "key": "foo"}
        assert result["conn_id"] == test_uuids[1]
        channels_info = result["channels_info"]["channels"]
        assert len(channels_info.keys()) == 2
        assert channels_info["a"]["total_users"] == 1
        assert channels_info["a"]["total_connections"] == 1
        assert channels_info["a"]["users"] == ["username"]
        assert channels_info["a"]["history"] == []
        assert result["channels_info"]["users"] == [
            {"state": {"bar": "baz", "key": "foo"}, "user": "username"}
        ]


@pytest.mark.usefixtures("cleanup_globals", "pyramid_config")
class TestUserStateViews(object):
    def test_bad_json(self, dummy_request, test_uuids):
        from channelstream.wsgi_views.server import user_state

        dummy_request.json_body = {}
        with pytest.raises(marshmallow.exceptions.ValidationError) as excinfo:
            user_state(dummy_request)
        assert excinfo.value.messages == {"user": ["Missing data for required field."]}

    def _connect_user(self, dummy_request, test_uuids):
        from channelstream.wsgi_views.server import connect

        dummy_request.json_body = {
            "username": "test",
            "conn_id": str(test_uuids[1]),
            "fresh_user_state": {"key": "foo"},
            "user_state": {"bar": "baz"},
            "state_public_keys": ["bar"],
            "channels": ["a", "aB"],
            "channel_configs": {"a": {"store_history": True, "history_size": 2}},
        }
        connect(dummy_request)

    def test_not_found_json(self, dummy_request, test_uuids):
        from channelstream.wsgi_views.server import user_state

        dummy_request.json_body = {"user": "blabla"}
        with pytest.raises(marshmallow.exceptions.ValidationError) as excinfo:
            user_state(dummy_request)
        assert excinfo.value.messages == {"user": ["Unknown user"]}

    def test_good_json(self, dummy_request, test_uuids):
        from channelstream.wsgi_views.server import user_state

        self._connect_user(dummy_request, test_uuids)
        dummy_request.json_body = {
            "user": "test",
            "user_state": {"bar": 2, "private": "im_private"},
            "state_public_keys": ["avatar", "bar"],
        }
        result = user_state(dummy_request)
        sorted_keys = sorted(["bar", "key", "private"])
        assert sorted_keys == sorted(result["user_state"].keys())
        assert result["user_state"]["private"] == "im_private"
        sorted_changed = sorted([x["key"] for x in result["changed_state"]])
        assert result["public_keys"] == ["avatar", "bar"]
        assert sorted_changed == sorted(["bar", "private"])

    def test_good_json_no_public_keys(self, dummy_request, test_uuids):
        from channelstream.wsgi_views.server import user_state

        self._connect_user(dummy_request, test_uuids)
        dummy_request.json_body = {
            "user": "test",
            "user_state": {"bar": 2, "private": "im_private"},
        }
        result = user_state(dummy_request)
        sorted_keys = sorted(["bar", "key", "private"])
        assert sorted_keys == sorted(result["user_state"].keys())
        assert result["user_state"]["private"] == "im_private"
        assert result["public_keys"] == ["bar"]
        sorted_changed = sorted([x["key"] for x in result["changed_state"]])
        assert sorted_changed == sorted(["bar", "private"])


@pytest.mark.usefixtures("cleanup_globals", "pyramid_config")
class TestSubscribeViews(object):
    def test_bad_json(self, dummy_request):
        from channelstream.wsgi_views.server import subscribe

        dummy_request.json_body = {}
        try:
            subscribe(dummy_request)
        except marshmallow.exceptions.ValidationError as exc:
            assert list(sorted(exc.messages.keys())) == ["channels", "conn_id"]

    def test_good_json(self, dummy_request, test_uuids):
        from channelstream.wsgi_views.server import connect, subscribe

        dummy_request.json_body = {
            "username": "test",
            "conn_id": str(test_uuids[1]),
            "fresh_user_state": {"key": "foo"},
            "user_state": {"bar": "baz"},
            "state_public_keys": ["bar"],
            "channels": ["a", "aB"],
            "channel_configs": {"a": {"store_history": True, "history_size": 2}},
        }
        connect(dummy_request)
        dummy_request.json_body = {
            "conn_id": str(test_uuids[1]),
            "channels": ["b"],
            "channel_configs": {
                "a": {"notify_presence": True},
                "b": {"notify_presence": True},
            },
        }
        result = subscribe(dummy_request)
        assert sorted(result["channels"]) == sorted(["a", "aB", "b"])
        assert result["channels_info"]["users"] == [
            {"state": {"bar": "baz", "key": "foo"}, "user": "test"}
        ]
        assert "a" in result["channels_info"]["channels"]
        assert "b" in result["channels_info"]["channels"]
        assert result["channels_info"]["channels"]["a"]["total_connections"] == 1
        assert result["channels_info"]["channels"]["a"]["total_users"] == 1
        assert result["channels_info"]["channels"]["a"]["history"] == []
        assert result["channels_info"]["channels"]["a"]["users"] == ["test"]


@pytest.mark.usefixtures("cleanup_globals", "pyramid_config")
class TestUnsubscribeViews(object):
    def test_bad_json(self, dummy_request, test_uuids):
        from channelstream.wsgi_views.server import unsubscribe

        dummy_request.json_body = {}
        try:
            unsubscribe(dummy_request)
        except marshmallow.exceptions.ValidationError as exc:
            assert list(sorted(exc.messages.keys())) == ["channels", "conn_id"]

    def test_good_json(self, dummy_request, test_uuids):
        from channelstream.wsgi_views.server import connect, unsubscribe

        dummy_request.json_body = {
            "username": "test",
            "conn_id": str(test_uuids[1]),
            "fresh_user_state": {"key": "foo"},
            "user_state": {"bar": "baz"},
            "state_public_keys": ["bar"],
            "channels": ["a", "aB", "aC"],
            "channel_configs": {"a": {"store_history": True, "history_size": 2}},
        }
        connect(dummy_request)
        dummy_request.json_body = {
            "conn_id": str(test_uuids[1]),
            "channels": ["aC", "a"],
        }
        result = unsubscribe(dummy_request)
        assert sorted(result["channels"]) == sorted(["aB"])

    def test_non_existing_channel(self, dummy_request, test_uuids):
        from channelstream.wsgi_views.server import connect, unsubscribe

        dummy_request.json_body = {
            "username": "test",
            "conn_id": str(test_uuids[1]),
            "fresh_user_state": {"key": "foo"},
            "user_state": {"bar": "baz"},
            "state_public_keys": ["bar"],
            "channels": ["a", "aB", "aC"],
            "channel_configs": {"a": {"store_history": True, "history_size": 2}},
        }

        connect(dummy_request)
        dummy_request.json_body = {"conn_id": str(test_uuids[1]), "channels": ["d"]}
        result = unsubscribe(dummy_request)
        assert sorted(result["channels"]) == sorted(["a", "aB", "aC"])

    def test_no_channels(self, dummy_request, test_uuids):
        from channelstream.wsgi_views.server import connect, unsubscribe

        dummy_request.json_body = {
            "username": "test",
            "conn_id": str(test_uuids[1]),
            "fresh_user_state": {"key": "foo"},
            "user_state": {"bar": "baz"},
            "state_public_keys": ["bar"],
            "channels": ["a"],
            "channel_configs": {"a": {"store_history": True, "history_size": 2}},
        }
        connect(dummy_request)
        dummy_request.json_body = {"conn_id": str(test_uuids[1]), "channels": ["a"]}
        result = unsubscribe(dummy_request)
        assert len(result["channels"]) == 0
        assert result["channels_info"]["users"] == []
        assert result["channels_info"]["channels"] == {}


@pytest.mark.usefixtures("cleanup_globals", "pyramid_config")
class TestInfoView(object):
    def test_empty_json(self, dummy_request):
        from channelstream.wsgi_views.server import info

        dummy_request.json_body = {}
        result = info(dummy_request)
        assert result["channels"] == {}
        assert result["users"] == []

    def test_subscribed_json(self, dummy_request, test_uuids):
        from channelstream.wsgi_views.server import connect, info

        dummy_request.json_body = {
            "username": "test1",
            "conn_id": str(test_uuids[1]),
            "fresh_user_state": {"key": "foo"},
            "user_state": {"bar": "baz"},
            "state_public_keys": ["bar"],
            "channels": ["a", "aB"],
            "channel_configs": {"a": {"store_history": True, "history_size": 2}},
        }
        connect(dummy_request)
        dummy_request.json_body = {
            "username": "test2",
            "conn_id": test_uuids[2],
            "fresh_user_state": {"key": "foo1"},
            "user_state": {"bar": "baz1"},
            "state_public_keys": ["key"],
            "channels": ["a", "c"],
            "channel_configs": {"c": {"store_history": True, "history_size": 2}},
        }
        connect(dummy_request)
        dummy_request.json_body = {}
        result = info(dummy_request)
        assert sorted(("a", "aB", "c")) == sorted(result["channels"].keys())
        assert result["users"]
        comp_a = sorted(result["channels"]["a"]["users"])
        comp_b = sorted(["test1", "test2"])
        assert comp_a == comp_b
        assert result["channels"]["a"]["total_users"] == 2
        assert result["channels"]["a"]["total_connections"] == 2
        assert result["channels"]["c"]["users"] == ["test2"]
        assert result["channels"]["c"]["total_users"] == 1
        assert result["channels"]["c"]["total_connections"] == 1
        assert result["channels"]["aB"]["users"] == ["test1"]
        comp_a = sorted(result["users"], key=lambda x: x["user"])
        comp_b = sorted(
            [
                {"state": {"bar": "baz", "key": "foo"}, "user": "test1"},
                {"state": {"bar": "baz1", "key": "foo1"}, "user": "test2"},
            ],
            key=lambda x: x["user"],
        )
        assert comp_a == comp_b
        dummy_request.body = "NOTEMPTY"
        dummy_request.json_body = {"info": {"channels": ["a"]}}
        result = info(dummy_request)
        assert "a" in result["channels"]
        assert "aB" not in result["channels"]

    def test_detailed_json(self, dummy_request, test_uuids):
        from channelstream.wsgi_views.server import connect, info, message

        dummy_request.json_body = {
            "username": "test1",
            "conn_id": str(test_uuids[1]),
            "fresh_user_state": {"key": "foo"},
            "user_state": {"bar": "baz", "private": "p1"},
            "state_public_keys": ["bar"],
            "channels": ["a", "aB", "c", "D"],
            "channel_configs": {"a": {"store_history": True, "history_size": 2}},
        }
        connect(dummy_request)
        dummy_request.json_body = [
            {
                "type": "message",
                "user": "test1",
                "channel": "a",
                "message": {"text": "test"},
            }
        ]
        message(dummy_request)
        gevent.sleep(0)
        dummy_request.body = "value"
        dummy_request.json_body = {
            "info": {
                "exclude_channels": ["c"],
                "include_history": False,
                "include_users": True,
                "return_public_state": True,
                "include_connections": True,
            }
        }
        result = info(dummy_request)
        assert sorted(result["channels"].keys()) == sorted(["a", "aB", "D"])
        assert "private" not in result["users"][0]["state"]
        assert len(result["channels"]["a"]["history"]) == 0


@pytest.mark.usefixtures("cleanup_globals", "pyramid_config")
class TestMessageViews(object):
    def test_empty_json(self, dummy_request):
        from channelstream.wsgi_views.server import message

        server_state = get_state()
        dummy_request.json_body = {}
        assert server_state.stats["total_unique_messages"] == 0

        with pytest.raises(marshmallow.exceptions.ValidationError) as excinfo:
            message(dummy_request)
        assert excinfo.value.messages == {"_schema": ["Invalid input type."]}

    def test_good_json_no_channel(self, dummy_request):
        from channelstream.wsgi_views.server import message

        server_state = get_state()
        channel = Channel("test")
        channel.store_history = True
        server_state.channels[channel.name] = channel
        msg_payload = {
            "type": "message",
            "user": "system",
            "channel": "test",
            "message": {"text": "test"},
        }

        dummy_request.json_body = [msg_payload]
        assert server_state.stats["total_unique_messages"] == 0
        assert len(channel.history) == 0
        message(dummy_request)
        # change context
        gevent.sleep(0)
        assert server_state.stats["total_unique_messages"] == 1
        assert len(channel.history) == 1
        msg = channel.history[0]
        assert msg["uuid"] is not None
        assert msg["user"] == msg_payload["user"]
        assert msg["message"] == msg_payload["message"]
        assert msg["type"] == msg_payload["type"]
        assert msg["channel"] == msg_payload["channel"]
        assert msg["timestamp"] is not None

    def test_catchup_messages(self, dummy_request):
        from channelstream.wsgi_views.server import message, connect

        server_state = get_state()
        dummy_request.json_body = {
            "username": "test1",
            "channels": ["test"],
            "channel_configs": {"test": {"store_history": True, "history_size": 2}},
        }
        connect(dummy_request)
        msg_payload = {
            "type": "message",
            "user": "system",
            "channel": "test",
            "message": {"text": "test3"},
        }
        dummy_request.json_body = [msg_payload]
        message(dummy_request)
        # add pm message to non-existing user
        wrong_user_msg_payload = {
            "type": "message",
            "user": "system",
            "channel": "test",
            "message": {"text": "test1"},
            "pm_users": ["test2"],
        }
        msg_payload = {
            "type": "message",
            "user": "system",
            "channel": "test",
            "message": {"text": "test2"},
            "pm_users": ["test1"],
        }
        dummy_request.json_body = [wrong_user_msg_payload, msg_payload]
        message(dummy_request)
        # change context
        gevent.sleep(0)
        connection = server_state.users["test1"].connections[0]
        messages = connection.get_catchup_messages()
        assert len(messages) == 2
        assert messages[0]["timestamp"] > connection.last_active
        assert messages[0]["message"]["text"] == "test3"
        assert messages[1]["timestamp"] > connection.last_active
        assert messages[1]["message"]["text"] == "test2"


@pytest.mark.usefixtures("cleanup_globals", "pyramid_config")
class TestMessageEditViews(object):
    def test_empty_json(self, dummy_request):
        from channelstream.wsgi_views.server import message

        dummy_request.json_body = {}
        with pytest.raises(marshmallow.exceptions.ValidationError) as excinfo:
            message(dummy_request)
        assert excinfo.value.messages == {"_schema": ["Invalid input type."]}

    def test_good_json_no_channel(self, dummy_request):
        from channelstream.wsgi_views.server import message, messages_patch

        server_state = get_state()
        channel = Channel("test")
        channel.store_history = True
        server_state.channels[channel.name] = channel
        msg_payload = {"user": "system", "channel": "test", "message": {"text": "test"}}
        dummy_request.json_body = [msg_payload]
        message(dummy_request)
        # change context
        gevent.sleep(0)
        msg = channel.history[0]
        assert msg["message"] == msg_payload["message"]
        edit_payload = {
            "uuid": msg["uuid"],
            "user": "edited_system",
            "channel": "test",
            "timestamp": "2010-01-01T01:01",
            "edited": "2010-01-01T01:02",
            "message": {"text": "edited_message"},
        }
        dummy_request.json_body = [edit_payload]
        response = messages_patch(dummy_request)[0]
        gevent.sleep(0)
        assert msg["user"] == response["user"]
        assert msg["message"] == response["message"]
        assert msg["edited"] == response["edited"]
        assert msg["timestamp"] == response["timestamp"]
        frame = channel.frames[0][1]
        assert id(frame) == id(msg)
        assert frame["user"] == response["user"]
        assert frame["message"] == response["message"]
        assert frame["edited"] == response["edited"]
        assert frame["timestamp"] == response["timestamp"]


class TestMessageDeleteViews(object):
    def test_empty_json(self, dummy_request):
        from channelstream.wsgi_views.server import messages_delete

        dummy_request.json_body = []
        result = messages_delete(dummy_request)
        assert result == []

    def test_good_json_no_channel(self, dummy_request):
        from channelstream.wsgi_views.server import message, messages_delete

        server_state = get_state()
        channel = Channel("test")
        channel.store_history = True
        server_state.channels[channel.name] = channel
        msg_payload = {"user": "system", "channel": "test", "message": {"text": "test"}}
        dummy_request.json_body = [msg_payload]
        message(dummy_request)
        # change context
        gevent.sleep(0)
        msg = channel.history[0]
        assert msg["message"] == msg_payload["message"]
        dummy_request.json_body = [{"uuid": str(msg["uuid"]), "channel": "test"}]
        response = messages_delete(dummy_request)
        gevent.sleep(0)
        assert response[0]["uuid"] == msg["uuid"]
        assert len(channel.history) == 0
        assert len(channel.frames) == 1
        assert channel.frames[0][1]["type"] == "message:delete"


@pytest.mark.usefixtures("cleanup_globals", "pyramid_config")
class TestChannelConfigView(object):
    def test_empty_json(self, dummy_request):
        from channelstream.wsgi_views.server import channel_config

        dummy_request.json_body = {}
        result = channel_config(dummy_request)
        assert result["channels"] == {}
        assert result["users"] == []

    def test_valid_json(self, dummy_request):
        from channelstream.wsgi_views.server import channel_config

        dummy_request.json_body = {
            "chanx1": {
                "notify_presence": True,
                "store_history": True,
                "history_size": 3,
                "broadcast_presence_with_user_lists": True,
                "notify_state": True,
                "store_frames": False,
            }
        }
        result = channel_config(dummy_request)
        channel_settings = result["channels"]["chanx1"]["settings"]
        assert channel_settings["notify_presence"] is True
        assert channel_settings["store_history"] is True
        assert channel_settings["history_size"] == 3
        assert channel_settings["broadcast_presence_with_user_lists"] is True
        assert channel_settings["notify_state"] is True
        assert channel_settings["store_frames"] is False



import logging
from pyramid.view import exception_view_config, forbidden_view_config
from pyramid.httpexceptions import HTTPForbidden, HTTPFound

log = logging.getLogger(__name__)


@exception_view_config(context="marshmallow.ValidationError", renderer="json")
def marshmallow_invalid_data(context, request):
    request.response.status = 422
    log.error("Request had incorrect payload")
    try:
        log.debug(context.messages)
        log.debug(request.text)
    except Exception as exc:
        log.warning(u"Can't convert debug messages: %s" % exc)
    return context.messages


@exception_view_config(context="channelstream.exceptions.BadSignature", renderer="json")
def signer_error(context, request):
    request.response.status = 401
    log.error("Request had incorrect signature")
    return {"request": "Bad Signature"}


@forbidden_view_config()
def unauthorized_handler(context, request, renderer="json"):
    if (
        request.matched_route
        and request.matched_route.pattern.startswith("/admin")
        or "api-explorer" in request.url
    ):
        url = request.route_url("admin_action", action="sign_in")
        return HTTPFound(url)
    return HTTPForbidden()



from datetime import datetime

from gevent.lock import RLock

STATS = {"started_on": datetime.utcnow()}
lock = RLock()


class State(object):
    def __init__(self):
        self.channels = {}
        self.connections = {}
        self.users = {}
        self.stats = {"total_messages": 0, "total_unique_messages": 0}
        self.lock = RLock()


STATES = {"0": State()}


def get_state(tenant_id="0"):
    """
    Grabs right state for specific tenant
    :param tenant_id:
    :return:
    """
    return STATES[tenant_id]



import copy
import uuid

import marshmallow
from itsdangerous import (
    TimestampSigner,
    BadSignature as itsBadSignature,
    BadTimeSignature,
)
from pyramid.renderers import render
from pyramid.settings import asbool

from channelstream.exceptions import BadSignature


def handle_cors(request):
    settings = request.registry.settings
    if not settings["allow_cors"]:
        request.response.headers.add("Access-Control-Allow-Origin", "*")
    else:
        origin = request.headers.get("Origin") or "<>"
        for domain in settings["allow_cors"]:
            if domain in origin:
                request.response.headers.add("Access-Control-Allow-Origin", "*")
                break

    request.response.headers.add("XDomainRequestAllowed", "1")
    request.response.headers.add(
        "Access-Control-Allow-Methods", "GET, POST, OPTIONS, PUT, DELETE, PATCH"
    )
    request.response.headers.add(
        "Access-Control-Allow-Headers",
        "Content-Type, Depth, User-Agent, "
        "X-File-Size, X-Requested-With, "
        "If-Modified-Since, X-File-Name, "
        "Cache-Control, Pragma, Origin, "
        "Connection, Referer, Cookie",
    )
    request.response.headers.add("Access-Control-Max-Age", "86400")


def swagger_ui_script_template(request, spec_route_name, **kwargs):
    """
    Generates the <script> code that bootstraps Swagger UI, it will be injected
    into index template
    :param request:
    :param spec_route_name:
    :return:
    """
    return render("templates/explorer.jinja2", value=(), request=request)


def uuid_from_string(str_uuid):
    try:
        return uuid.UUID(str_uuid)
    except (ValueError, AttributeError):
        raise marshmallow.ValidationError("Wrong UUID format")


def process_catchup(m):
    copied = copy.deepcopy(m)
    copied["catchup"] = True
    copied.pop("pm_users", None)
    copied.pop("exclude_users", None)
    copied.pop("no_history", None)
    return copied


def set_config_types(config):
    """
    convert raw config values to proper types
    :param config:
    :return:
    """
    config = copy.deepcopy(config)
    config["debug"] = asbool(config["debug"])
    config["port"] = int(config["port"])
    config["validate_requests"] = asbool(config["validate_requests"])
    config["enforce_https"] = asbool(config["enforce_https"])
    if not config["cookie_secret"]:
        config["cookie_secret"] = str(uuid.uuid4())

    if config["http_scheme"] not in ["http", "https"]:
        config["http_scheme"] = ""

    for key in ["allow_posting_from", "allow_cors"]:
        if not config[key]:
            continue
        # if those keys are strings from ini convert to lists of individual values
        if isinstance(config[key], str):
            try:
                listed = [ip.strip() for ip in config[key].split(",") if ip.strip()]
                config[key] = listed
            except ValueError:
                pass
    return config


class DefaultSigner:
    def __init__(self, secret, **kwargs):
        self.signer = TimestampSigner(secret, **kwargs)

    def unsign(self, secret, **kwargs):
        try:
            self.signer.unsign(secret, **kwargs)
        except (itsBadSignature, BadTimeSignature):
            raise BadSignature(f"Signature {secret} does not match")



import logging

from pyramid.security import Allow, Everyone, ALL_PERMISSIONS

log = logging.getLogger(__name__)


class RequestBasicChallenge(Exception):
    pass


def is_allowed_ip(addr, config):
    if "0.0.0.0" in config["allow_posting_from"]:
        return True
    return addr in config["allow_posting_from"]


class APIFactory:
    def __init__(self, request):
        self.__acl__ = []
        config = request.registry.settings
        req_url_secret = request.params.get("secret")
        req_secret = request.headers.get("x-channelstream-secret", req_url_secret)

        addr = request.environ["REMOTE_ADDR"]
        if not is_allowed_ip(addr, config):
            log.warning("IP: {} is not whitelisted".format(addr))
            return

        if req_secret:
            max_age = 60 if config["validate_requests"] else None
            request.registry.signature_checker.unsign(req_secret, max_age=max_age)
        else:
            return
        self.__acl__ = [(Allow, Everyone, ALL_PERMISSIONS)]


class AdminAuthFactory(object):
    def __init__(self, request):
        self.__acl__ = []
        user = request.authenticated_userid
        if user:
            self.__acl__ = [(Allow, Everyone, ALL_PERMISSIONS)]
        else:
            # try the API auth factory too for server requests
            context_obj = APIFactory(request)
            self.__acl__ = context_obj.__acl__



from gevent import monkey

monkey.patch_all()

import copy
import pytest

from webtest import TestApp
from channelstream.cli import SHARED_DEFAULTS
from channelstream.cli.start import RoutingApplication


def gen_signature(secret):
    from itsdangerous import TimestampSigner

    signer = TimestampSigner(secret)
    return signer.sign("channelstream")


@pytest.mark.usefixtures("cleanup_globals")
class TestServerConfigView(object):
    def test_defaults(self):
        config = copy.deepcopy(SHARED_DEFAULTS)
        application = RoutingApplication(config)
        app = TestApp(application, extra_environ={"REMOTE_ADDR": "127.0.0.1"})
        app.get("/", status=404)
        app.post("/connect", status=403)

    def test_enforce_https_wrong(self):
        config = copy.deepcopy(SHARED_DEFAULTS)
        config["enforce_https"] = True
        application = RoutingApplication(config)
        app = TestApp(application, extra_environ={"REMOTE_ADDR": "127.0.0.1"})
        app.get("/", status=406)
        app.post("/connect", status=406)

    def test_enforce_https_proper(self):
        config = copy.deepcopy(SHARED_DEFAULTS)
        config["enforce_https"] = True
        application = RoutingApplication(config)
        app = TestApp(
            application,
            extra_environ={"REMOTE_ADDR": "127.0.0.1", "wsgi.url_scheme": "https"},
        )
        app.get("/", status=404)
        app.post("/connect", status=403)

    def test_connect_empty(self):
        config = copy.deepcopy(SHARED_DEFAULTS)
        application = RoutingApplication(config)
        app = TestApp(application, extra_environ={"REMOTE_ADDR": "127.0.0.1"})
        secret_headers = {
            "x-channelstream-secret": gen_signature(SHARED_DEFAULTS["secret"]),
            "Content-Type": "application/json",
        }
        app.post_json("/connect", params={}, headers=secret_headers, status=422)

    def test_connect_minimal(self):
        config = copy.deepcopy(SHARED_DEFAULTS)
        application = RoutingApplication(config)
        app = TestApp(application, extra_environ={"REMOTE_ADDR": "127.0.0.1"})
        secret_headers = {
            "x-channelstream-secret": gen_signature(SHARED_DEFAULTS["secret"]),
            "Content-Type": "application/json",
        }
        payload = {
            "username": "user_123",
            # where user should be subscribed
            "channels": ["a"],
        }
        result = app.post_json("/connect", params=payload, headers=secret_headers)
        assert "conn_id" in result.json
        assert "state" in result.json
        assert "username" in result.json
        assert "public_state" in result.json
        assert result.json["channels"] == ["a"]
        assert "channels_info" in result.json
        assert "a" in result.json["channels"]
        channel_a = result.json["channels_info"]["channels"]["a"]
        assert "uuid" in channel_a
        assert "name" in channel_a
        assert "long_name" in channel_a
        assert "last_active" in channel_a
        assert channel_a["total_connections"] == 1
        assert channel_a["total_users"] == 1
        assert channel_a["users"] == ["user_123"]
        assert channel_a["total_connections"] == 1
        assert channel_a["settings"]["notify_presence"] is False
        assert channel_a["settings"]["store_history"] is False
        assert channel_a["settings"]["history_size"] == 10
        assert channel_a["settings"]["broadcast_presence_with_user_lists"] is False
        assert channel_a["settings"]["notify_presence"] is False
        assert channel_a["settings"]["notify_state"] is False
        assert channel_a["settings"]["store_frames"] is True
        assert channel_a["settings"]["store_frames"] is True
        assert len(channel_a["history"]) == 0
        assert len(result.json["channels_info"]["users"]) == 1
        assert result.json["channels_info"]["users"][0] == {
            "user": "user_123",
            "state": {},
        }

    def test_connect_good(self):
        config = copy.deepcopy(SHARED_DEFAULTS)
        application = RoutingApplication(config)
        app = TestApp(application, extra_environ={"REMOTE_ADDR": "127.0.0.1"})
        secret_headers = {
            "x-channelstream-secret": gen_signature(SHARED_DEFAULTS["secret"]),
            "Content-Type": "application/json",
        }
        CHANNEL_CONFIGS = {
            "pub_chan": {
                "notify_presence": True,
                "notify_state": True,
                "store_history": True,
                "history_size": 10,
                "broadcast_presence_with_user_lists": True,
            },
            "notify": {
                "store_history": True,
                "history_size": 50,
                "notify_presence": True,
            },
        }

        payload = {
            "username": "user_123",
            # where user should be subscribed
            "channels": ["a", "b"],
            # what default state should be set when user is created on channelstream end
            "fresh_user_state": {
                "status": "active",
                "private": "is private",
                "bar": 1,
                "bool": True,
            },
            # update state to this values if user object already exists
            "user_state": {},
            # what state keys should be visible to other users
            "state_public_keys": ["status", "bar", "color"],
            # return only public state in response
            "info": {"return_public_state": True},
            # set chanel configurations if channels don't exist yet
            "channel_configs": CHANNEL_CONFIGS,
        }
        app.post_json("/connect", params=payload, headers=secret_headers)

    def test_message_minimal(self):
        config = copy.deepcopy(SHARED_DEFAULTS)
        application = RoutingApplication(config)
        app = TestApp(application, extra_environ={"REMOTE_ADDR": "127.0.0.1"})
        secret_headers = {
            "x-channelstream-secret": gen_signature(SHARED_DEFAULTS["secret"]),
            "Content-Type": "application/json",
        }
        payload = {
            "type": "message",
            "user": "system",
            "channel": "some_chan",
            "message": {"text": "my_text"},
        }
        result = app.post_json("/message", params=[payload], headers=secret_headers)
        message = result.json[0]
        assert len(message["pm_users"]) == 0
        assert len(message["exclude_users"]) == 0
        assert message["channel"] == "some_chan"
        assert message["no_history"] is False
        assert message["edited"] is None
        assert "uuid" in message
        assert "timestamp" in message
        assert message["user"] == "system"
        assert message["message"] == {"text": "my_text"}



from pyramid.httpexceptions import HTTPNotAcceptable


def handle_new_request(event):
    settings = event.request.registry.settings
    event.request.handle_cors()
    if settings["enforce_https"] and event.request.scheme != "https":
        raise HTTPNotAcceptable("SSL is required")

    if settings["http_scheme"]:
        event.request.scheme = settings["http_scheme"]



from __future__ import absolute_import

import datetime
import decimal
import functools
import json
import uuid


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, complex):
            return [obj.real, obj.imag]
        elif isinstance(obj, datetime.datetime):
            r = obj.isoformat()
            if r.endswith("+00:00"):
                r = r[:-6] + "Z"
            return r
        elif isinstance(obj, uuid.UUID):
            return str(obj)
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, decimal.Decimal):
            return str(obj)
        elif isinstance(obj, datetime.time):
            r = obj.isoformat()
            if obj.microsecond:
                r = r[:12]
            return r
        elif isinstance(obj, set):
            return list(obj)
        elif hasattr(obj, "__json__"):
            if callable(obj.__json__):
                return obj.__json__()
            else:
                return obj.__json__
        else:
            raise NotImplementedError


load = json.load
loads = json.loads
dump = json.load
dumps = functools.partial(json.dumps, indent=4, cls=ComplexEncoder)



from setuptools import setup, find_packages
from os import path
from io import open

here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

requires = [
    "gevent>=1.3.7",
    "ws4py>=0.5.1",
    "marshmallow==2.18.0,<3.0.0",
    "python-dateutil>=2.7.5",
    "pyramid>=1.9.2",
    "pyramid_jinja2>=2.7",
    "pyramid_apispec==0.3.3",
    "itsdangerous==1.1.0",
    "requests>=2.20.0",
]

setup(
    name="channelstream",
    version="0.7.1",
    description="Websocket server supporting channels/users communication",
    keywords="websockets async communication chat notification django flask pyramid",
    url="https://github.com/Channelstream/channelstream",
    classifiers=[
        "Framework :: Django",
        "Framework :: Flask",
        "Framework :: Pyramid",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Communications :: Chat",
        "Topic :: Communications :: Conferencing",
    ],
    author="Marcin Lulek",
    author_email="info@webreactor.eu",
    long_description=long_description,
    license="BSD",
    long_description_content_type="text/markdown",
    zip_safe=True,
    packages=find_packages(),
    include_package_data=True,
    install_requires=requires,
    python_requires=">=3.6",
    setup_requires=["pytest-runner"],
    extras_require={
        "dev": ["coverage", "pytest", "pyramid", "tox", "mock", "webtest", "mypy"],
        "lint": ["black"],
    },
    entry_points={
        "console_scripts": [
            "channelstream = channelstream.cli.start:main",
            "channelstream_utils = channelstream.cli.utils:main",
        ]
    },
)



from gevent import monkey

monkey.patch_all()

import pytest
from datetime import datetime, timedelta
from gevent.queue import Queue
from channelstream.server_state import get_state
import channelstream.gc
from channelstream.channel import Channel
from channelstream.connection import Connection
from channelstream.user import User


@pytest.mark.usefixtures("cleanup_globals", "test_uuids")
class TestChannel(object):
    def test_create_defaults(self):
        channel = Channel("test", long_name="long name")
        assert channel.name == "test"
        assert channel.long_name == "long name"
        assert channel.connections == {}
        assert channel.notify_presence is False
        assert channel.broadcast_presence_with_user_lists is False
        assert channel.salvageable is False
        assert channel.store_history is False
        assert channel.history_size == 10
        assert channel.history == []

    def test_repr(self):
        channel = Channel("test", long_name="long name")
        assert repr(channel) == "<Channel: test, connections:0>"

    @pytest.mark.parametrize(
        "prop, value",
        [
            ("notify_presence", True),
            ("store_history", 6),
            ("history_size", 42),
            ("broadcast_presence_with_user_lists", True),
        ],
    )
    def test_create_set_config(self, prop, value):
        channel_config = {prop: value}
        channel = Channel("test", channel_config=channel_config)

        assert getattr(channel, prop) == value

    def test_add_connection(self, test_uuids):
        connection = Connection("test_user", conn_id=test_uuids[1])
        channel = Channel("test")
        channel.add_connection(connection)
        assert len(channel.connections["test_user"]) == 1
        assert "test_user" in channel.connections
        assert connection in channel.connections["test_user"]
        assert repr(channel) == "<Channel: test, connections:1>"

    def test_remove_connection(self, test_uuids):
        connection = Connection("test_user", conn_id=test_uuids[1])
        connection2 = Connection("test_user2", conn_id=test_uuids[2])
        connection3 = Connection("test_user", conn_id=test_uuids[3])
        channel = Channel("test")
        channel.add_connection(connection)
        channel.add_connection(connection2)
        channel.remove_connection(connection)
        assert "test_user" not in channel.connections
        assert len(channel.connections["test_user2"]) == 1
        channel.add_connection(connection)
        channel.add_connection(connection3)
        channel.remove_connection(connection)
        assert len(channel.connections["test_user"]) == 1

    def test_remove_non_existant_connection(self, test_uuids):
        channel = Channel("test")
        connection = Connection("test_user", conn_id=test_uuids[1])
        channel.remove_connection(connection)
        assert "test_user" not in channel.connections

    def test_remove_connection_w_presence(self, test_uuids):
        server_state = get_state()
        user = User("test_user")
        server_state.users[user.username] = user
        connection = Connection("test_user", conn_id=test_uuids[1])
        user.add_connection(connection)
        config = {"notify_presence": True, "broadcast_presence_with_user_lists": True}
        channel = Channel("test", channel_config=config)
        channel.add_connection(connection)
        channel.remove_connection(connection)

    def test_add_connection_w_presence(self, test_uuids):
        server_state = get_state()
        user = User("test_user")
        server_state.users[user.username] = user
        connection = Connection("test_user", conn_id=test_uuids[1])
        user.add_connection(connection)
        config = {"notify_presence": True, "broadcast_presence_with_user_lists": True}
        channel = Channel("test", channel_config=config)
        channel.add_connection(connection)
        assert len(channel.connections["test_user"]) == 1
        assert "test_user" in channel.connections
        assert connection in channel.connections["test_user"]

    def test_presence_message(self, test_uuids):
        user = User("test_user")
        connection = Connection("test_user", conn_id=test_uuids[1])
        user.add_connection(connection)
        channel = Channel("test")
        channel.add_connection(connection)
        payload = channel.send_notify_presence_info("test_user", "join")
        assert payload["user"] == "test_user"
        assert payload["message"]["action"] == "join"
        assert payload["type"] == "presence"
        assert payload["channel"] == "test"
        assert len(payload["users"]) == 0

    def test_presence_message_w_users(self, test_uuids):
        server_state = get_state()
        user = User("test_user")
        user.state_from_dict({"key": "1", "key2": "2"})
        user.state_public_keys = ["key2"]
        server_state.users[user.username] = user
        connection = Connection("test_user", conn_id=test_uuids[1])
        user.add_connection(connection)
        user2 = User("test_user2")
        user2.state_from_dict({"key": "1", "key2": "2"})
        server_state.users[user2.username] = user2
        connection2 = Connection("test_user2", conn_id=test_uuids[2])
        user2.add_connection(connection2)
        config = {"notify_presence": True, "broadcast_presence_with_user_lists": True}
        channel = Channel("test", channel_config=config)
        channel.add_connection(connection)
        channel.add_connection(connection2)
        payload = channel.send_notify_presence_info("test_user", "join")
        assert len(payload["users"]) == 2
        sorted_users = sorted(payload["users"], key=lambda x: x["user"])
        assert sorted_users == [
            {"state": {"key2": "2"}, "user": "test_user"},
            {"state": {}, "user": "test_user2"},
        ]

    def test_history(self):
        config = {"store_history": True, "history_size": 3}
        channel = Channel("test", long_name="long name", channel_config=config)
        channel.add_message(
            {
                "channel": "test",
                "message": "test1",
                "type": "message",
                "no_history": False,
                "pm_users": [],
                "exclude_users": [],
            }
        )
        channel.add_message(
            {
                "channel": "test",
                "message": "test2",
                "type": "message",
                "no_history": False,
                "pm_users": [],
                "exclude_users": [],
            }
        )
        channel.add_message(
            {
                "channel": "test",
                "message": "test3",
                "type": "message",
                "no_history": False,
                "pm_users": [],
                "exclude_users": [],
            }
        )
        channel.add_message(
            {
                "channel": "test",
                "message": "test4",
                "type": "message",
                "no_history": True,
                "pm_users": [],
                "exclude_users": [],
            }
        )
        channel.add_message(
            {
                "channel": "test",
                "message": "test5",
                "type": "message",
                "no_history": False,
                "pm_users": [],
                "exclude_users": [],
            }
        )

        assert len(channel.history) == 3
        assert channel.history == [
            {
                "channel": "test",
                "message": "test2",
                "type": "message",
                "no_history": False,
                "pm_users": [],
                "exclude_users": [],
            },
            {
                "channel": "test",
                "message": "test3",
                "type": "message",
                "no_history": False,
                "pm_users": [],
                "exclude_users": [],
            },
            {
                "channel": "test",
                "message": "test5",
                "type": "message",
                "no_history": False,
                "pm_users": [],
                "exclude_users": [],
            },
        ]

    def test_user_state(self, test_uuids):
        user = User("test_user")
        changed = user.state_from_dict({"key": "1", "key2": "2"})
        user.state_public_keys = ["key2"]
        connection = Connection("test_user", conn_id=test_uuids[1])
        user.add_connection(connection)
        channel = Channel("test")
        channel.add_connection(connection)
        payload = channel.send_user_state(user, changed)
        assert payload["user"] == "test_user"
        assert payload["message"]["state"] == {"key2": "2"}
        assert payload["message"]["changed"] == [{"key": "key2", "value": "2"}]
        assert payload["type"] == "user_state_change"
        assert payload["channel"] == "test"

    def test_user_single_assignment(self, test_uuids):
        server_state = get_state()
        user = User("test_user")
        connection = Connection("test_user", conn_id=test_uuids[1])
        user.add_connection(connection)
        channel = Channel("test")
        server_state.channels[channel.name] = channel
        channel.add_connection(connection)
        assert [channel] == user.get_channels()

    def test_user_multi_assignment(self, test_uuids):
        server_state = get_state()
        user = User("test_user")
        connection = Connection("test_user", conn_id=test_uuids[1])
        connection2 = Connection("test_user", conn_id=test_uuids[2])
        connection3 = Connection("test_user", conn_id=test_uuids[3])
        user.add_connection(connection)
        user.add_connection(connection2)
        user.add_connection(connection3)
        channel = Channel("test")
        channel2 = Channel("test2")
        server_state.channels[channel.name] = channel
        server_state.channels[channel2.name] = channel2
        channel.add_connection(connection)
        channel.add_connection(connection2)
        channel2.add_connection(connection3)
        assert ["test", "test2"] == sorted([c.name for c in user.get_channels()])


@pytest.mark.usefixtures("cleanup_globals")
class TestConnection(object):
    def test_create_defaults(self, test_uuids):
        now = datetime.utcnow()
        connection = Connection("test", test_uuids[1])
        assert connection.username == "test"
        assert now <= connection.last_active
        assert connection.socket is None
        assert connection.queue is None
        assert connection.id == test_uuids[1]

    def test_mark_for_gc(self, test_uuids):
        long_time_ago = datetime.utcnow() - timedelta(days=50)
        connection = Connection("test", test_uuids[1])
        connection.mark_for_gc()
        assert connection.last_active < long_time_ago

    def test_message(self, test_uuids):
        connection = Connection("test", test_uuids[1])
        connection.queue = Queue()
        connection.add_message({"message": "test"})
        assert connection.queue.get() == [{"message": "test"}]

    def test_heartbeat(self, test_uuids):
        connection = Connection("test", test_uuids[1])
        connection.queue = Queue()
        connection.heartbeat()
        assert connection.queue.get() == []


class TestUser(object):
    def test_create_defaults(self):
        user = User("test_user")
        user.state_from_dict({"key": "1", "key2": "2"})
        user.state_public_keys = ["key2"]
        assert repr(user) == "<User:test_user, connections:0>"
        assert sorted(user.state.items()) == sorted({"key": "1", "key2": "2"}.items())
        assert user.public_state == {"key2": "2"}

    def test_messages(self, test_uuids):
        user = User("test_user")
        connection = Connection("test_user", conn_id=test_uuids[1])
        connection.queue = Queue()
        connection2 = Connection("test_user", conn_id=test_uuids[2])
        connection2.queue = Queue()
        user.add_connection(connection)
        user.add_connection(connection2)
        user.add_message(
            {
                "type": "message",
                "no_history": False,
                "pm_users": [],
                "exclude_users": [],
            }
        )
        assert len(user.connections) == 2
        assert len(user.connections[0].queue.get()) == 1
        assert len(user.connections[1].queue.get()) == 1


@pytest.mark.usefixtures("cleanup_globals")
class TestGC(object):
    def test_gc_connections_active(self, test_uuids):
        server_state = get_state()
        channel = Channel("test")
        server_state.channels[channel.name] = channel
        channel2 = Channel("test2")
        server_state.channels[channel2.name] = channel2
        user = User("test_user")
        server_state.users[user.username] = user
        user2 = User("test_user2")
        server_state.users[user2.username] = user2
        connection = Connection("test_user", test_uuids[1])
        server_state.connections[connection.id] = connection
        connection2 = Connection("test_user", test_uuids[2])
        server_state.connections[connection2.id] = connection2
        connection3 = Connection("test_user2", test_uuids[3])
        server_state.connections[connection3.id] = connection3
        connection4 = Connection("test_user2", test_uuids[4])
        server_state.connections[connection4.id] = connection4
        user.add_connection(connection)
        user.add_connection(connection2)
        channel.add_connection(connection)
        channel.add_connection(connection2)
        user2.add_connection(connection3)
        user2.add_connection(connection4)
        channel2.add_connection(connection3)
        channel2.add_connection(connection4)
        channelstream.gc.gc_conns()
        conns = server_state.channels["test"].connections["test_user"]
        assert len(conns) == 2
        assert len(server_state.connections.items()) == 4
        conns = server_state.channels["test2"].connections["test_user2"]
        assert len(conns) == 2
        assert len(user.connections) == 2
        assert len(user2.connections) == 2
        assert sorted(channel.connections.keys()) == ["test_user"]
        assert sorted(channel2.connections.keys()) == ["test_user2"]

    def test_gc_connections_collecting(self, test_uuids):
        server_state = get_state()
        channel = Channel("test")
        server_state.channels[channel.name] = channel
        channel2 = Channel("test2")
        server_state.channels[channel2.name] = channel2
        user = User("test_user")
        server_state.users[user.username] = user
        user2 = User("test_user2")
        server_state.users[user2.username] = user2
        connection = Connection("test_user", test_uuids[1])
        server_state.connections[connection.id] = connection
        connection2 = Connection("test_user", test_uuids[2])
        connection2.mark_for_gc()
        server_state.connections[connection2.id] = connection2
        connection3 = Connection("test_user2", test_uuids[3])
        connection3.mark_for_gc()
        server_state.connections[connection3.id] = connection3
        connection4 = Connection("test_user2", test_uuids[4])
        server_state.connections[connection4.id] = connection4
        user.add_connection(connection)
        user.add_connection(connection2)
        channel.add_connection(connection)
        channel.add_connection(connection2)
        user2.add_connection(connection3)
        user2.add_connection(connection4)
        channel2.add_connection(connection3)
        channel2.add_connection(connection4)
        channelstream.gc.gc_conns()
        assert len(server_state.connections.items()) == 2
        conns = server_state.channels["test"].connections["test_user"]
        assert len(conns) == 1
        assert conns == [connection]
        conns = server_state.channels["test2"].connections["test_user2"]
        assert len(conns) == 1
        assert conns == [connection4]
        assert len(user.connections) == 1
        assert len(user2.connections) == 1
        connection.mark_for_gc()
        connection4.mark_for_gc()
        channelstream.gc.gc_conns()
        assert "test_user" not in server_state.channels["test"].connections
        assert "test_user2" not in server_state.channels["test2"].connections
        assert len(server_state.channels["test"].connections.items()) == 0
        assert len(server_state.channels["test2"].connections.items()) == 0

    def test_users_active(self):
        server_state = get_state()
        user = User("test_user")
        server_state.users[user.username] = user
        user2 = User("test_user2")
        server_state.users[user2.username] = user2
        channelstream.gc.gc_users()
        assert len(server_state.users.items()) == 2
        user.last_active -= timedelta(days=2)
        channelstream.gc.gc_users()
        assert len(server_state.users.items()) == 1
